//
// Fl_Table -- A table widget
//
// Copyright 2002 by Greg Ercolano.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
// Please report all bugs and problems to "erco@seriss.com".
//
// TODO:
//    o Auto scroll during dragged selection
//    o Keyboard navigation (up/down/left/right arrow)
//    o Row headings (only column headings supported currently)
//

#include <FL/fl_draw.H>
#include "Fl_Table.H"

// Scroll display so 'row' is at top
void Fl_Table::top_row(int row)
{
    if ( row < 0 || row >= rows() ) return;
    double newtop = row * rowh;
    if ( newtop > vscrollbar->maximum() ) 
        { newtop = vscrollbar->maximum(); }
    vscrollbar->Fl_Slider::value(newtop);
    table_scrolled();
    redraw();
}

// Ctor
Fl_Table::Fl_Table(int X, int Y, int W, int H, const char *l) : Fl_Group(X,Y,W,H,l)
{
    _rows       = 0;
    _cols       = 0;
    rowh        = 18;
    toprow      = 0;
    botrow      = 0;
    leftcol     = 0;
    rightcol    = 0;
    tablew      = 0;
    tableh      = 0;
    last_cursor = FL_CURSOR_DEFAULT;
    _when       = FL_WHEN_RELEASE;

    box(FL_THIN_DOWN_FRAME);

    tablebox = new Fl_Box(x(), y(), w(), h());
    tablebox->box(FL_NO_BOX);

    vscrollbar = new Fl_Scrollbar(x()+w()-17, y(), 17, h()-17);
    vscrollbar->type(FL_VERTICAL);
    vscrollbar->callback(scroll_cb, (void*)this);

    hscrollbar = new Fl_Scrollbar(x(), y()+h()-17, w(), 17);
    hscrollbar->type(FL_HORIZONTAL);
    hscrollbar->callback(scroll_cb, (void*)this);

    table_resized();

    end();
}

// Dtor
Fl_Table::~Fl_Table()
{
}

// Set height of a row.
//     Parent must call redraw().
//     Currently all rows are the same height.
//
void Fl_Table::row_height(int row, int new_h)
{
    if ( row < 0 || row >= _rows ) { return; }
    // TODO: Allow rows to be different sizes
    rowh = new_h;
}

// Set width of a row.
//     Parent must call redraw().
//
void Fl_Table::col_width(int col, int new_w)
{
    if ( col < 0 || col >= _cols ) { return; }
    colw[col] = new_w;
}

// Return row/col clamped to reality
int Fl_Table::row_col_clamp(int &R, int &C)
{
    int clamped = 0;
    if ( R < 0 )      { R = 0;       clamped = 1; }
    if ( R >= _rows ) { R = _rows-1; clamped = 1; }
    if ( C < 0 )      { C = 0;       clamped = 1; }
    if ( C >= _cols ) { C = _cols-1; clamped = 1; }
    return(clamped);
}

// Return bounding region for given context
void Fl_Table::get_bounds(TableContext context, int &X, int &Y, int &W, int &H)
{
    switch ( context )
    {
        case CONTEXT_COL_HEADER:

	    // Column header clipping.
	    X = tablebox->x();
	    Y = tablebox->y() - rowh;
	    W = tablebox->w();
	    H = rowh;
	    return;

	case CONTEXT_TABLE:

	    // Table window clipping.
	    //     If table window touches any Fl_Table widget edge, 
	    //     compensate for the 1 pixel border all around its inside.
	    //
	    X = tablebox->x();
	    Y = tablebox->y();
	    W = tablebox->w();
	    H = tablebox->h();
	    return;

	// TODO: Add other contexts..
	default:
	    // fprintf(stderr, "get_bounds(): context %d unimplemented\n", context);
	    return;
    }
    //NOTREACHED
}

// Find row/col beneath cursor
//
//    Returns R/C and context.
//    Also returns resizeflag, if mouse is hovered over a resize boundary.
//
Fl_Table::TableContext Fl_Table::cursor2rowcol(int &R, int &C, ResizeFlag &resizeflag)
{
    R = C = 0;
    resizeflag = RESIZE_NONE;
    
    int X, Y, W, H;

    // Event inside a scrollbar?
    //    Then event is not in table context.
    //
    if ( ( vscrollbar->visible() && Fl::event_inside(vscrollbar) ) || 
         ( hscrollbar->visible() && Fl::event_inside(hscrollbar) ) )
    {
        return(CONTEXT_NONE);
    }
    
    // Column header?
    if ( col_header() )
    {
        get_bounds(CONTEXT_COL_HEADER, X, Y, W, H);

	// Inside a column heading?
	if ( Fl::event_inside(X, Y, W, H) )
	{
	    // Scan visible columns until found
	    for ( C = leftcol; C <= rightcol; C++ )
	    {
		find_cell(CONTEXT_COL_HEADER, 0, C, X, Y, W, H);
		if ( Fl::event_x() >= X && Fl::event_x() < (X+W) )
		{
		    // Found column?
		    //     If cursor over resize boundary, and resize enabled,
		    //     enable the appropriate resize flag.
		    //
		    if ( col_resize() )
		    {
		        if ( Fl::event_x() <= (X+3-0) ) { resizeflag = RESIZE_COL_LEFT; }
		        if ( Fl::event_x() >= (X+W-3) ) { resizeflag = RESIZE_COL_RIGHT; }
		    }
		    return(CONTEXT_COL_HEADER);
		}
	    }

	    // Must be in column header dead zone
	    return(CONTEXT_NONE);
	}
    }

    // Mouse somewhere in table?
    //     Scan visible r/c's until we find it.
    //
    for ( R = toprow; R <= botrow; R++ )
    {
	for ( C = leftcol; C <= rightcol; C++ )
	{
	    find_cell(CONTEXT_CELL, R, C, X, Y, W, H);
	    if ( Fl::event_inside(X, Y, W, H) )
	        { return(CONTEXT_CELL); }	// found it
	}
    }

    // Somewhere else in table
    return(CONTEXT_NONE);
}

// Find X/Y/W/H for cell at R/C
//     If R or C are out of range, returns -1 
//     with X/Y/W/H set to zero.
//
int Fl_Table::find_cell(TableContext context, int R, int C, int &X, int &Y, int &W, int &H)
{
    if ( row_col_clamp(R, C) )
	{ X=Y=W=H=0; return(-1); }

    for ( int c=X=0; c<C; c++ )
	{ X += colw[c]; }

    X = X - hscrollbar->value() + tablebox->x();
    Y = R * rowh - vscrollbar->value() + tablebox->y();
    W = colw[C];
    H = rowh;

    switch ( context )
    {
	case CONTEXT_COL_HEADER:
	    Y = tablebox->y() - rowh;		// force vertical position above tablebox
	    return(0);

	case CONTEXT_ROW_HEADER:
	    X = ( colw.size() == 0 ) ? 		// force horiz position left of tablebox
	        tablebox->x() : (tablebox->x() - colw[0]);
	    return(0);

	case CONTEXT_CELL:
	    return(0);

	// TODO -- HANDLE OTHER CONTEXTS
	default:
	    // fprintf(stderr, "find_cell: unknown context %d\n", (int)context);
	    return(-1);
    }
    //NOTREACHED
}

// Return table window size
void Fl_Table::table_window_size(int &X, int &Y, int &W, int &H)
{
    X=x(); Y=y(); H=h(); W=w();

    // Trim window if scrollbars visible
    if ( vscrollbar->visible() ) { W -= 17; }
    if ( hscrollbar->visible() ) { H -= 17; }

    // Trim window if header enabled
    if ( col_header() )
    {
        Y += rowh;
	H -= rowh;
    }
}

// Recalculate internals after a scroll.
//
//    Call this if table has been scrolled or resized.
//    Does not handle redraw().
//    TODO: Assumes tablebox has already been recalculated.
//
void Fl_Table::table_scrolled()
{
    if ( _rows == 0 || _cols == 0 )
        return;

    // Top row
    int voff = vscrollbar->value();
    toprow = voff / rowh;
    if ( toprow >= _rows ) toprow = _rows - 1;

    // Bottom row
    voff = vscrollbar->value();
    botrow = ( voff + tablebox->h() ) / rowh;
    if ( botrow >= _rows ) botrow = _rows - 1;

    // Left column
    int x, hoff = hscrollbar->value();
    for ( leftcol=x=0; leftcol < _cols; leftcol++ )
    {
        x += colw[leftcol];
        if ( x >= hoff ) { x -= colw[leftcol]; break; }
    }
    leftcol = ( leftcol >= _cols ) ? (leftcol - 1) : leftcol;

    // Right column
    //    Work with data left over from leftcol calculation
    //
    hoff = hscrollbar->value() + w();
    for ( rightcol=leftcol; rightcol < _cols; rightcol++ )
    {
        x += colw[rightcol];
        if ( x >= hoff ) { break; }
    }
    rightcol = ( rightcol >= _cols ) ? (rightcol - 1) : rightcol;
}

// Table resized: recalc internal data
//    Call this whenever the window is resized.
//    Recalculates the scrollbar sizes.
//    Makes no assumptions about any pre-initialized data.
//
void Fl_Table::table_resized()
{
    // Height
    tableh = rowh * _rows;

    // Width
    for ( int col=tablew=0; col < (int)colw.size(); col++ )
        { tablew += colw[col]; }
    
    //
    // Make scroll bars disappear if window large enough
    //

    // First pass: can hide via window size?
    int hidev = ( tableh - h() < 0 ),
	hideh = ( tablew - w() < 0 );

    // Second pass: Check for interference
    if ( !hideh & hidev ) { hidev = ( ( tableh - h() + 17 ) < 0 ); } 
    if ( !hidev & hideh ) { hideh = ( ( tablew - w() + 17 ) < 0 ); }

    // Determine scrollbar visibility
    if ( hidev ) { vscrollbar->hide(); } else { vscrollbar->show(); }
    if ( hideh ) { hscrollbar->hide(); } else { hscrollbar->show(); }

    //
    // Get the table's physical clipping window.
    //
    //    Do this AFTER hide/showing scrollbars
    //    since if they're shown, they affect 
    //    the size of the table window.
    //
    int twx, twy, tww, twh;
    table_window_size(twx, twy, tww, twh);
    tablebox->resize(twx, twy, tww, twh);
    tablebox->redraw();

    // Recalc scrollbar sizes
    //    Clamp scrollbar value() after a resize.
    //    Resize scrollbars to enforce a constant trough width after a window resize.
    //
    {
        float vscrolltab = ( tableh == 0 || twh > tableh ) ? 1 : (float)twh / tableh;
        float hscrolltab = ( tablew == 0 || tww > tablew ) ? 1 : (float)tww / tablew;

	vscrollbar->bounds(0, tableh-twh);
	vscrollbar->precision(10);
	vscrollbar->slider_size(vscrolltab);
	vscrollbar->resize(x()+w()-17, y(), 17, h() - ((hideh)?0:17) );
	vscrollbar->Fl_Valuator::value(vscrollbar->clamp(vscrollbar->value()));	

	hscrollbar->bounds(0, tablew-tww);
	hscrollbar->precision(10);
        hscrollbar->slider_size(hscrolltab);
	hscrollbar->resize(x(), y()+h()-17, w() - ((hidev)?0:17), 17);
	hscrollbar->Fl_Valuator::value(hscrollbar->clamp(hscrollbar->value()));
    }

    // Tell FLTK child widgets were resized
    Fl_Group::init_sizes();

    // Recalc top/bot/left/right
    table_scrolled();

    // Redraw everything
    redraw();
}

// Someone moved a scrollbar
void Fl_Table::scroll_cb(Fl_Widget*w, void *data)
{
    Fl_Table *o = (Fl_Table*)data;
    o->table_scrolled();
    o->redraw();
}

// Set number of rows
void Fl_Table::rows(int val)
{
    _rows = val;
    table_resized();
}

// Set number of cols
void Fl_Table::cols(int val)
{
    _cols = val;
    {
	int default_w = ( colw.size() > 0 ) ? colw.back() : 80;
	while ( val > (int)colw.size() ) { colw.push_back(default_w); }	// enlarge
	while ( val < (int)colw.size() ) { colw.pop_back(); }		// shrink
    }

    table_resized();
}

// Change mouse cursor to different type
void Fl_Table::change_cursor(Fl_Cursor newcursor)
{
    if ( newcursor != last_cursor )
    {
	fl_cursor(newcursor, FL_BLACK, FL_WHITE);
	last_cursor = newcursor;
    }
}

// #include "eventnames.h"

// Handle FLTK events
int Fl_Table::handle(int event)
{
    // fprintf(stderr, "** EVENT: %s: ", eventnames[event]);

    static int resizing_col = -1,			// column being dragged
	       dragging_x = -1,				// starting x position for drag
	       push_row = -1,				// FL_PUSH on this row (used by FL_RELEASE)
	       push_col = -1;				// FL_PUSH on this col (used by FL_RELEASE)

    static TableContext push_context = CONTEXT_NONE;	// FL_PUSH context

    int ret = Fl_Group::handle(event);	// let FLTK group handle events first

    // Which row/column are we over?
    int R, C;  				// row/column being worked on
    ResizeFlag resizeflag;		// which resizing area are we over? (0=none)
    TableContext context = cursor2rowcol(R, C, resizeflag);

    switch ( event )
    {
	case FL_PUSH:

	    // Save these for later when button released.
	    //    Do this for all mouse push buttons.
	    //
	    push_context = context;
	    push_row = R;
	    push_col = C;

	    if ( context == CONTEXT_COL_HEADER && Fl::event_button() == 1 && resizeflag )
	    {
		// Start resize if left mouse clicked on column border.
		//
		//    "ret = 1" ensures we get drag events from now on.
		//    (C-1) is used if mouse is over the left hand side of the cell,
		//    so that we resize the next column over to the left.
		//
		resizing_col = ( resizeflag & RESIZE_COL_LEFT ) ? C-1 : C; 
		dragging_x = Fl::event_x(); 
		ret = 1;
	    }

	    break;

        case FL_DRAG:

	    // Dragging column?
	    //
	    //    Let user drag even /outside/ the row/col widget.
	    //    Don't allow column width smaller than 1.
	    //    Continue to show FL_CURSOR_WE at all times during drag.
	    //
	    if ( resizing_col > -1 )
	    {
		int offset = dragging_x - Fl::event_x();
		colw[resizing_col] -= offset;
		if ( colw[resizing_col] < 1 )
		    { colw[resizing_col] = 1; }
		dragging_x = Fl::event_x();
		table_resized();
		change_cursor(FL_CURSOR_WE);
		ret = 1;
	    }
	    break;

	case FL_RELEASE:
	    switch ( context )
	    {
		case CONTEXT_COL_HEADER:	// release on header
		case CONTEXT_CELL:		// release on a cell

		    // FL_RELEASE on same row/col we FL_PUSH'ed?
		    //    Then do the user's callback with context.
		    //
		    if ( _when & (FL_WHEN_RELEASE | FL_WHEN_RELEASE_ALWAYS ) &&	// when qualifies?
		         push_context == context && 				// same context?
		         push_row == R && 					// same row FL_PUSH'ed?
			 push_col == C && 					// same col FL_PUSH'ed?
			 resizing_col == -1 )					// not resizing a column
		    {
		        callback(context, R, C);	// don't use Fl_Table::callback(),
							// or parent class can't override
		    }
		    break;

		default:
		    break;
	    }

	    if ( Fl::event_button() == 1 )
	    {
		change_cursor(FL_CURSOR_DEFAULT);
		resizing_col = -1;
		ret = 1;
	    }
	    break;

	case FL_MOVE:
	    if ( context == CONTEXT_COL_HEADER && 	// in column header?
	         resizeflag )				// and resize enabled + near resize boundary?
		{ change_cursor(FL_CURSOR_WE); }	// show resize cursor
	    else
		{ change_cursor(FL_CURSOR_DEFAULT); }	// normal cursor
	    ret = 1;
	    break;

	default:
	    change_cursor(FL_CURSOR_DEFAULT);
	    break;
    }

    return(ret);
}

// Resize FLTK override
//     Handle resize events if user resizes parent window.
//
void Fl_Table::resize(int X, int Y, int W, int H)
{
    // Tell group to resize, and recalc our own widget as well
    Fl_Group::resize(X, Y, W, H);
    table_resized();
}

// Draw the entire Fl_Table
//    Override the draw() routine to draw the table.
//    Then tell the group to draw over us.
//
void Fl_Table::draw()
{
    int X,Y,W,H;

    draw_cell(CONTEXT_NEWPAGE);		// let user's drawing routine prepare for new page

    // Draw column headers
    if ( col_header() )
    {
	// Draw column header *above* 'tablebox'; compensate y clipping.
	fl_push_clip(tablebox->x(), tablebox->y()-rowh, tablebox->w(), rowh);

	for ( int c = leftcol; c <= rightcol; c++ )
	{
	    find_cell(CONTEXT_COL_HEADER, 0, c, X, Y, W, H);
	    draw_cell(CONTEXT_COL_HEADER, 0, c, X, Y, W, H);
	}
	fl_pop_clip();
    }

    // Draw all cells.
    //    This includes cells partially obscured off edges of table.
    //
    int cx, cy, cw, ch;
    get_bounds(CONTEXT_TABLE, cx, cy, cw, ch);
    fl_push_clip(cx, cy, cw, ch);
    {
	for ( int r = toprow; r <= botrow; r++ )
	{
	    for ( int c = leftcol; c <= rightcol; c++ )
	    {
		find_cell(CONTEXT_CELL, r, c, X, Y, W, H);
		draw_cell(CONTEXT_CELL, r, c, X, Y, W, H);
	    }
	}
    }
    fl_pop_clip();

    // Dead zone right of table visible?
    if ( cw > tablew )
    {
        // Fill in with a rectangle
        int X = cx + cw - ( cw - tablew ), Y = y(),
	    W = cw - tablew, H = h();
        fl_color(color());
	fl_rectf(X,Y,W,H);
    }

    // Dead zone below table visible?
    if ( ch > tableh )
    {
        // Fill in with a rectangle
        int X = cx, Y = y() + ch - ( ch - tableh ),
	    W = w(), H = ch - tableh;

	// TODO
	//     Shouldn't have to do this. We probably need another tablebox-like
	//     object that has the dimensions of the ENTIRE table, including headers.
	//
	if ( col_header() ) { Y += rowh; } 
        fl_color(color());
	fl_rectf(X,Y,W,H);
    }

    // Dead zone between scrollbars?
    if ( vscrollbar->visible() && hscrollbar->visible() )
    {
        fl_color(color());
	fl_rectf(x()+w()-17, y()+h()-17, 17, 17);
    }

    // Draw group over the above
    //
    //     Group's background is transparent (a frame)
    //     so this really just draws the thin frame around the table.
    //
    Fl_Group::draw();
}
